
//////////////////////////////////////////////////////////////////////////////
//
//  xinoutd.C
//
//  $Log$
//
//	Written by Shaun Terry, Copyright 1995
//
//////////////////////////////////////////////////////////////////////////////

#ifndef lint
static const char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

extern "C" int write(int, char *, int);
extern "C" int read(int, char *, int);
extern "C" int getdtablesize();
extern "C" int close(int);

extern int errno;

#define SERVER

#include "xinout.h"
#include "socket.h"

int cbRead(int);

int fFortune = FALSE;

int fDebug;

void cbAlarm();

const char *usage = "xinoutd [ [ -fortune | -cmd command ] -delay minutes ] [ -debug ]\n";

int sock;
fd_set rm;
int tbllen;

int delay = 15;
char *cmd = 0;

int main(int argc, char **argv)
{
	char *port = getenv("XINOUTPORT");
	int portNum = 0;
	if (port) portNum = atoi(port);

    if (fFortune = getParamFlag(argc, argv, "-fortune"))
			cmd = "/usr/games/fortune -s";

    getParamString(argc, argv, "-cmd", cmd);
    getParamInt(argc, argv, "-delay", delay);

	if (getParamFlag(argc, argv, "-help")) {
		printf(usage);
		return 0;
	}
	fDebug = getParamFlag(argc, argv, "-debug");

	Socket s;

	// Try listening on the specified port
	if (portNum > 0) {
		if ((sock = s.hostListen(portNum, 5)) < 0) {
			perror("Listen: ");
			return(1);
		}
	}
	// Try listening on the service name
	else if ((sock = s.hostListen("xinout", 5)) > 0) {
		;	// OK
	}
	// Try listening on the default port name
	else if ((sock = s.hostListen(7232, 5)) > 0) {
		;	// OK
	}
	else { perror("Listen: "); return 1; }

	if (sock < 0) { perror("listen"); return 1; }

    FD_ZERO(&rm);
    FD_SET(sock, &rm);

	tbllen = getdtablesize();

	fd_set tmprm = rm;

	signal(SIGALRM, (SIG_PF) cbAlarm);
	if (cmd && delay > 0) alarm(delay * 60);

	while (1) {
		if (select(tbllen, &tmprm, 0, 0, 0 ) > 0) {
			if (FD_ISSET(sock, &tmprm)) {
				int newSock = accept(sock, 0, 0);
				FD_SET(newSock, &rm);
				if (fDebug) printf("Accepted %d\n", newSock);
			}

			for (int i=0; i < tbllen; ++i) {
				if (i == sock) continue;
				if (FD_ISSET(i, &tmprm)) {
					if (cbRead(i) < 0) {
						if (fDebug) printf("Closing socket %d\n", i);
						FD_CLR(i, &rm);
						close(i);
					}
				}
			}
		}
		else if (errno != EINTR) perror("select");

		tmprm = rm;
	}

	return 0;
}

void cbAlarm()
{
	char buff[1024];

	if (fDebug) printf("Alarm [%s]\n", cmd);
	FILE *fp = popen(cmd, "r");
	if (!fp) return;

	int c=0, x=0;
	for (int i=0; i < sizeof(buff)-1; ++i) {
		c = fgetc(fp);
		// Label widget don't like tabs
		if (c == '\t' && fFortune) continue;
		if (c == EOF) break;
		buff[x++] = c;
	}
	buff[x] = '\0';
	if (buff[strlen(buff)-1] == '\n') buff[strlen(buff)-1] = '\0';
	if (fDebug) printf("%s\n", buff);

	pclose(fp);

	if (strlen(buff)) {

		inout_msg_hdr hdr;

		// Re-send the message to all other xinout's
		hdr.type = SET_FORTUNE;
		strcpy(hdr.who, "");
		hdr.in_or_out = 'x';
		hdr.textlen = strlen(buff)+1;
		for (int i=0; i < tbllen; ++i) {
			if (i == ::sock) continue;
			if (FD_ISSET(i, &rm)) {
				write(i, (char *) &hdr, sizeof(hdr));
				if (hdr.textlen) write(i, buff, hdr.textlen);
			}
		}
	}

	if (cmd && delay > 0) alarm(delay * 60);
}

int cbRead(int sock)
{
	if (fDebug) printf("Incoming message\n");
	inout_msg_hdr hdr;
	char *buf = 0;

	if (read(sock, (char *) &hdr, sizeof(hdr)) <= 0) return -1;
	if (hdr.textlen > 0) {
		buf = new char[hdr.textlen];
		read(sock, buf, hdr.textlen);
	}

	if (fDebug) {
		printf("Type = %d  who = [%s] in_or_out = %c\n", hdr.type, hdr.who,
				hdr.in_or_out);
		printf("Text = [%s]\n", (buf) ? buf : "");
	}

	// Re-send the message to all other xinout's
	for (int i=0; i < tbllen; ++i) {
		if (i == ::sock || i == sock) continue;
		if (FD_ISSET(i, &rm)) {
			write(i, (char *) &hdr, sizeof(hdr));
			if (hdr.textlen) write(i, buf, hdr.textlen);
		}
	}

	if (buf) delete [] buf;

	return 0;
}

